// キャラクタの動作などに関するクラス群
// Pacman、Ghostともにcharacterクラスを継承している。

// 継承を設定するための関数。
// childにたいしてparentのprototypeをひも付けている。
var inherits = function(childCtor,parentCtor){
  Object.setPrototypeOf(childCtor.prototype,parentCtor.prototype);
};

// キャラクター全体に関するクラス。
// x,yの座標、動く方向、周囲のマスの状況などをメンバ変数として持つ。
var character = function(start,move){
  this.x = start[0];
  this.y = start[1];
  this.move = move;
  this.currentPosition = [this.x,this.y];
  this.surroundings = {};
};
character.prototype.update = function(){
  // キャラクターの位置を更新する。
  // 現在キャラクターが向いている方向に1マス分動く動作を行う。
  if(this.state != "dead"){
    var next = this.nextPosition();
    this.x = next[0];
    this.y = next[1];
    this.currentPosition = [this.x,this.y];
  }
};
character.prototype.nextPosition = function(){
  var x = this.x + direction[this.move][0];
  var y = this.y + direction[this.move][1];
  if(this.x >= mapSizeX - 1 && this.move == "right") x = 0;
  if(this.x <= 0 && this.move == "left") x = mapSizeX - 1;
  if(this.y >= mapSizeY - 1 && this.move == "up") y = 0;
  if(this.y <= 0 && this.move == "down") y = mapSizeY - 1;
  return [x,y];
};
character.prototype.legalMove = function(map){
  if(this.y >= mapSizeY - 1){
    this.surroundings["up"] = map[this.y - 1][this.x];
    this.surroundings["down"] = map[0][this.x];
  }else if(this.y <= 0){
    this.surroundings["up"] = map[mapSizeY-1][this.y];
    this.surroundings["down"] = map[this.y + 1][this.x];
  }else{
    this.surroundings["up"] = map[this.y - 1][this.x];
    this.surroundings["down"] = map[this.y + 1][this.x];
  }
  if(this.x >= mapSizeX - 1){
    this.surroundings["right"] = map[this.y][0];
    this.surroundings["left"] = map[this.y][this.x - 1];
  }else if(this.x <= 0){
    this.surroundings["right"] = map[this.y][this.x + 1];
    this.surroundings["left"] = map[this.y][mapSizeX-1];
  }else{
    this.surroundings["right"] = map[this.y][this.x + 1];
    this.surroundings["left"] = map[this.y][this.x - 1];
  }
  this.surroundings[direct_pair[this.move]] = 1;
  var surroundings = this.surroundings;
  var legalMove = Object.keys(direction).filter(function(d){
    return surroundings[d] != 1;
  });
  return legalMove;
}
character.prototype.randomMove = function(map){
  legalMove = this.legalMove(map);
  this.move = goodShuffle(legalMove)[0];
}

var Pacman = function(start, move){
  character.call(this, start,move);
  this.monteCarloData = {};
  this.monteCarloTreeRoot;
  this.moveMethod = "tree";
  this.cowardice = 3;
};
inherits(Pacman, character);
Pacman.prototype.directionUpdate = function(map,ghost){
  this.randomMove(map);
  if(this.moveMethod == "primitive"){
    this.monteCarloMove(ghost,map);
  }else if(this.moveMethod == "tree"){
    this.monteCarloTree(ghost,map);
  }
};
Pacman.prototype.monteCarloMove = function(gho,ma){
  for(var i=0; i<MaxTry; i++){
    var buf = playout(this,gho,ma);//playoutは[score,[intersection,move]]を返す
    if(buf[1].length == 0){
      this.monteCarloData["firstIntersection"] = "none";
      break;
    }
    if(this.monteCarloData["firstIntersection"] != buf[1][0]){
      this.monteCarloData["firstIntersection"] = buf[1][0];
      this.monteCarloData["up"] = [];
      this.monteCarloData["down"] = [];
      this.monteCarloData["left"] = [];
      this.monteCarloData["right"] = [];
    }
    this.monteCarloData[buf[1][1]].push(buf[2] * buf[0] / fullScore);
  }
  if(this.monteCarloData["firstIntersection"] != "none"){
    var scoreArr = [];
    var data = this.monteCarloData;
    Object.keys(this.monteCarloData).forEach(function(d){
      if(d != "firstIntersection" && data[d].length != 0){
        scoreArr.push(arrAverage(data[d]));
      }else{
        scoreArr.push(0);
      }
    });
    if(this.monteCarloData["firstIntersection"] == this.y + "," + this.x){
      var idx = scoreArr.indexOf(Math.max.apply(null,scoreArr));
      if(Object.keys(this.monteCarloData)[idx] != "firstIntersection")  this.move = Object.keys(this.monteCarloData)[idx];
    }
  }
}

var Ghost = function(start, move, state){
  character.call(this, start, move);
  this.state = state;
  this.stateCount = 0;
  this.aggression = 5;
  this.aggressionMeter = 0;
};
inherits(Ghost, character);
Ghost.prototype.stateUpdate = function(){
  if(this.state != "normal"){
    this.stateCount -= 1;
    if(this.stateCount == 0){
      this.state = "normal"
    }
  }
};
Ghost.prototype.directionUpdate = function(map,pacman){
  this.randomMove(map);
  var pacPosition = pacman.y + "," + pacman.x;
  var ghostPosition = this.y + "," + this.x;
  if(routeMap[pacPosition][ghostPosition] != undefined){
    if(Math.random() < this.aggression/10 &&
       routeMap[pacPosition][ghostPosition].cost < 10&&
       this.aggressionMeter == 0){
      this.aggressionMeter = 10;
    }else if(this.aggressionMeter > 0){
      this.dijkstra(pacman);
      this.aggressionMeter -= 1
      if(this.aggressionMeter == 0) this.aggressionMeter -= 10;
    }else if (this.aggressionMeter < 0) {
      this.aggressionMeter += 1;
    }
  }
};
Ghost.prototype.dijkstra = function(pacman){
  var pacPosition = pacman.y + "," + pacman.x;
  var ghostPosition = this.y + "," + this.x;
  if(routeMap[pacPosition][ghostPosition] != undefined){
    var directionBuf = routeMap[pacPosition][ghostPosition].from;
  }else{
    var directionBuf = "none";
  }
  if(directionBuf != "none"){
    var nextMoveX = parseInt(directionBuf.split(",")[1]) - this.x;
    var nextMoveY = parseInt(directionBuf.split(",")[0]) - this.y;
    if(Math.abs(nextMoveX) < 2 && Math.abs(nextMoveY) < 2){
      var nextMove = Object.keys(direction).filter(function(d){
        return (direction[d][0] == nextMoveX && direction[d][1] == nextMoveY);
      });
      this.move = nextMove;
    }
  }
};
