// AIによる意思決定は各交差点(intersections)において行われる。
// intersectionsの判定は、そのセルにおいてキャラクターが動くことの出来る方向の数によって決定されうるが、毎回計算をするのは、やや手間である。
// 全ての交差点をゲーム開始時にグラフにしておくことにした。
// 各交差点における移動可能方向、およびその方向にどの交差点があるのかをハッシュにした。
// intersections["y,x"]の形で移動可能方向。
// graph["y,x"]の形で接続された交差点を取れる。
// また、最終的には「次に意思決定を必要とする交差点はどこか」（例えば、曲がり角では意思決定を必要としない）
// ということがわかるようにしておきたい。

function makeGraph(){
  Object.keys(intersections).forEach(function(d){
    graph[d] = searchNearIntersection(d);
  });
};

function searchNearIntersection(intersectionFocusOn){
  var node = {};
  splited = intersectionFocusOn.split(",");
  var x = parseInt(splited[1]);
  var y = parseInt(splited[0]);
  intersections[intersectionFocusOn].forEach(function(d){
    var move = d;
    var cell = nextCell(intersectionFocusOn,move);
    var next;
    while(intersections[next = cell.next().value] == undefined){
    };
    node[move] = next;
  });
  return node;
};

function* nextCell(intersectionFocusOn,move){
  splited = intersectionFocusOn.split(",");
  var x = parseInt(splited[1]);
  var y = parseInt(splited[0]);
  while(1){
    x += direction[move][0];
    y += direction[move][1];
    if(y > mapSizeY - 1){
      y = 0;
    }else if(y < 0){
      y = mapSizeY - 1;
    }
    if(x > mapSizeX - 1){
      x = 0;
    }else if(x < 0){
      x = mapSizeX - 1;
    }
    yield y + "," + x;
  }
}

function intersectionLookUp(map){
  var intersections = {};
  map.forEach(function(d, i){
    var y = i;
    d.forEach(function(d, i){
      var x = i;
      if(map[y][x] != 1){
        var direction = {};
        if(y >= mapSizeY - 1){
          direction["up"] = map[y - 1][x];
          direction["down"] = map[0][x];
        }else if(y <= 0){
          direction["up"] = map[mapSizeY - 1][x];
          direction["down"] = map[y + 1][x];
        }else{
          direction["up"] = map[y - 1][x];
          direction["down"] = map[y + 1][x];
        }
        if(x >= mapSizeX - 1){
          direction["right"] = map[y][0];
          direction["left"] = map[y][x - 1];
        }else if(x <= 0){
          direction["right"] = map[y][x + 1];
          direction["left"] = map[y][mapSizeX - 1];
        }else{
          direction["right"] = map[y][x + 1];
          direction["left"] = map[y][x - 1];
        }
        var legalDirection = Object.keys(direction).filter(function(d){return direction[d] != 1});
        if(legalDirection.toString() != 'right,left' &&
           legalDirection.toString() != 'up,down'){
           intersections[[y,x].toString()] = legalDirection;
        }
      }
    });
  });
  return intersections;
};
