//　モンテカルロ木探索についての関数群。
// monteCarloTreeはパックマンの意思決定を行うためのメソッドである。
// ゴーストとパックマンの現時点の位置情報からモンテカルロ木探索を行う。
//
// Nodeはモンテカルロ木探索のノードを構成するクラスであり、このメソッドとして木をたどる関数などが定義されている。

Pacman.prototype.monteCarloTree = function(gho,ma){
  var nextIntersection = playout(this,gho,ma)[1][0];
  if(nextIntersection == undefined) return;
  if(this.monteCarloTreeRoot != undefined){
    if(nextIntersection != this.monteCarloTreeRoot.cell) var root = new Node(null, null, nextIntersection);
    else var root = this.monteCarloTreeRoot;
  }else{
    var root = new Node(null, null, nextIntersection);
  }

  for(var i = 0; i < MaxTry; i++){
    var node = root;
    var route = [];
    while (node.untriedMoves.length == 0 && node.childNodes.length != 0){
      node = node.selectChild();
      route.push(node.move);
    }
    if (node.untriedMoves.length !== 0 && node.try > 20){
      node = node.makeChildNode();
      route.push(node.move)
    }
    var playoutResult = playout(this,gho,ma,route);
    var clear = playoutResult[2] * playoutResult[0] / fullScore;
    node.backpropagation(clear);
  }
  var result = root.childNodes.map(function (n) {return n.try;});
  this.monteCarloTreeRoot = root;
  if(root.cell == this.y + "," + this.x){
    var maxNode = root.childNodes[result.indexOf(Math.max.apply(null,result))];
    this.move = maxNode.move;
  }
}

function Node(parentNode, move, cell) {
  this.parentNode = parentNode;
  this.move = move;
  this.childNodes = [];
  this.score = 0;
  this.try = 0;
  this.cell = cell;
  this.untriedMoves = [].concat(intersections[cell]);
}
// 子ノードの中からUCTの最も高いノードを選ぶメソッド
Node.prototype.selectChild = function(){
  var totalTry = this.try;
  var UCTvalues = this.childNodes.map(function(d){
    var c = Math.sqrt(2);
    return d.score / d.try +
           c * Math.sqrt(Math.log(totalTry) / d.try);
  });
  // console.log("this",this);
  // console.log("UCTvalues",UCTvalues);
  return this.childNodes[UCTvalues.indexOf(Math.max.apply(null, UCTvalues))];
}
// ノードに未試行の手がある場合、それについて、新しい子ノードを作る。
// この時、untriedMovesからその手を削除する。
Node.prototype.makeChildNode = function () {
  var move = goodShuffle(this.untriedMoves)[0];
  var childIntersection = searchNearIntersection(this.cell)[move];
  // console.log(childIntersection);
  var childNode = new Node(this, move, childIntersection);
  this.childNodes.push(childNode);
  this.untriedMoves.splice(this.untriedMoves.indexOf(move),1);
  // console.log(this.untriedMoves);
  return childNode;
}
// プレイアウト結果の親への伝搬を行うメソッド
Node.prototype.backpropagation = function(clear){
  for (var node=this; node !== null; node = node.parentNode){
    node.update(clear);
  }
}
Node.prototype.update = function(clear){
  this.score += clear;
  this.try += 1;
}

function playout(pac,gho,ma,route){
  if(route == undefined){
    route = [];
  }
  var map = mapCopy(ma);
  var pacman = new Pacman([pac.x,pac.y],pac.move);
  var ghost = [];
  gho.forEach(function(d,i){
    ghost.push(new Ghost([d.x,d.y],d.move,d.state));
    ghost[i].aggression = pac.cowardice;
  });

  var finish = 0;
  var value = 0;
  var clear = 0;
  var firstIntersection = [];

  for(var i=0;i<=playoutCycle;i++){
    pacman.randomMove(map);
    if(intersections[pacman.y + "," + pacman.x] != undefined &&
       intersections[pacman.y + "," + pacman.x].length >= 2){
      if(firstIntersection.length == 0){
        firstIntersection.push(pacman.y + "," + pacman.x);
        firstIntersection.push(pacman.move);
      }
      if(route.length != 0){
        pacman.move = route[0];
        route.splice(0,1);
      }
    }
    ghost.forEach(function(d,i){
      d.directionUpdate(map,pacman);
      if(CollisionDetection(pacman,d)){
        if(d.state == "ijike"){
          ghost.splice(i,1);
          if(ghost.length == 0){
            clear = 1;
            finish = 1;
          };
        }else{
          finish = 1;
        }
      }else{
        d.update();
        d.stateUpdate();
      }
    });
    if(finish == 1){
      break;
    }
    pacman.update();
    if(map[pacman.y][pacman.x] == 2){
      value += 10;
      if(value == fullScore){
        clear = 1;
        finish = 1;
      }
    }else if(map[pacman.y][pacman.x] == 3){
      ghost.forEach(function(d){
        d.state = "ijike";
        d.stateCount = ijikeCount;
      });
    }
    map[pacman.y][pacman.x] = 0;
    if(i == playoutCycle){
      clear = 1;
      finish = 1;
    }
  }

  return [value,firstIntersection,clear]
};
